
require('react');
require('graphql');
require('lodash');
require('subscriptions-transport-ws');

const { generateApolloClient } = require("@deep-foundation/hasura/client");
const { DeepClient } = require('@deep-foundation/deeplinks/imports/client');
const { minilinks, Link } = require('@deep-foundation/deeplinks/imports/minilinks');

const f = async () => {
  const apolloClient = generateApolloClient({
    path: '3006-deepfoundation-dev-hp7ied6ymka.ws-us47.gitpod.io/gql', // <<= HERE PATH TO UPDATE
    ssl: true,
    // admin token in prealpha deep secret key
    token: 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJodHRwczovL2hhc3VyYS5pby9qd3QvY2xhaW1zIjp7IngtaGFzdXJhLWFsbG93ZWQtcm9sZXMiOlsibGluayJdLCJ4LWhhc3VyYS1kZWZhdWx0LXJvbGUiOiJsaW5rIiwieC1oYXN1cmEtdXNlci1pZCI6IjI0OCJ9LCJpYXQiOjE2NTQ4MTY4Njl9.iDboJXGc9a08e9h75Vm_G9_EjsARyXNGm5KUZWzz2e4',
  });
  const deep = new DeepClient({ apolloClient, linkId: 248 });
  deep.linkId // actualUserId
  const Type = await deep.id('@deep-foundation/core', 'Type') // Type type link id
  const Package = await deep.id('@deep-foundation/core', 'Package') // Package type link Id
  const Contain = await deep.id('@deep-foundation/core', 'Contain') // Contain type link id
  const Any = await deep.id('@deep-foundation/core', 'Any');

  console.log({ Type, Package, Contain, Any });

  const { data: [{ id: packageId }] } = await deep.insert({
    type_id: Package,
    string: { data: { value: `@deep-foundation/messaging` } },
    in: {
      data: {
        type_id: Contain,
        from_id: deep.linkId
      }
    },
  });

  console.log({ packageId });

  // insert type Reply
  const { data: [{ id: Reply }] } = await deep.insert({
    // Reply = from Any to Any
    type_id: Type,
    from_id: Any,
    to_id: Any,
    in: {
      data: {
        // nest into package as 'Reply' in Contain's tree under package
        type_id: Contain,
        from_id: packageId, // before created package
        string: { data: { value: 'Reply' } },
      }
    },
  });

  console.log({ Reply });

  const Value = await deep.id('@deep-foundation/core', 'Value'); // Value type link id
  const StringId = await deep.id('@deep-foundation/core', 'String'); // String link id - use it as symbol

  console.log({ Value, StringId });

  // insert type Message
  const { data: [{ id: Message }] } = await deep.insert({
    // Message just dot
    type_id: Type,
    in: {
      data: {
        // nest into package as 'Message' in Contain's tree under package
        type_id: Contain,
        from_id: packageId, // before created package
        string: { data: { value: 'Message' } },
      }
    },
    out: {
      data: {
        // Message can have .string as .value of link.
        type_id: Value,
        to_id: StringId,
      }
    },
  });

  console.log({ Message });

  // insert type Author
  const { data: [{ id: Author }] } = await deep.insert({
    // Reply = from Any to Any
    type_id: Type,
    from_id: Any,
    to_id: Any,
    in: {
      data: {
        // nest into package as 'Author' in Contain's tree under package
        type_id: Contain,
        from_id: packageId, // before created package
        string: { data: { value: 'Author' } },
      }
    },
  });

  console.log({ Author });

  // insert type Join
  const { data: [{ id: Join }] } = await deep.insert({
    // Reply = from Any to Any
    type_id: Type,
    from_id: Any,
    to_id: Any,
    in: {
      data: {
        // nest into package as 'Join' in Contain's tree under package
        type_id: Contain,
        from_id: packageId, // before created package
        string: { data: { value: 'Join' } },
      }
    },
  });

  console.log({ Join });

  const SyncTextFile = await deep.id('@deep-foundation/core', 'SyncTextFile');
  const dockerSupportsJs = await deep.id('@deep-foundation/core', 'dockerSupportsJs');
  const Handler = await deep.id('@deep-foundation/core', 'Handler');
  const HandleInsert = await deep.id('@deep-foundation/core', 'HandleInsert');

  console.log({ SyncTextFile, dockerSupportsJs, Handler, HandleInsert });

  const { data: [{ id: handlerId }] } = await deep.insert({
    type_id: SyncTextFile,
    in: {
      data: {
        from_id: dockerSupportsJs,
        type_id: Handler,
        in: {
          data: {
            type_id: HandleInsert,
            from_id: Join,
          }
        },
      }
    },
    string: {
      data: {
        value: `async (arg) => {
      const deep = arg.deep;
      const Query = await deep.id('@deep-foundation/core', 'Query');
      return { Query };
    }` }
    },
    in: {
      data: {
        // nest into package as 'Join' in Contain's tree under package
        type_id: Contain,
        from_id: packageId, // before created package
        string: { data: { value: 'joinInsertHandler' } },
      }
    },
  });

  console.log({ handlerId });

  const guest0 = await deep.guest();
  const deepGuest0 = new DeepClient({ deep, ...guest0 });

  // insert first message in actual user
  await deep.insert({
    type_id: Message,
    string: { data: { value: 'test message' } },
  });
  // insert first message in actual user
  const { data: [{ id: messageId1 }] } = await deep.insert({
    type_id: Message,
    string: { data: { value: 'first message' } },
    out: {
      data: {
        type_id: Reply,
        to_id: deepGuest0.linkId,
        // join deepGuest0 from here in down by messagingTree
        out: {
          data: {
            type_id: Join,
            to_id: deepGuest0.linkId,
          }
        }
      }
    },
  });

  console.log({ guest0, messageId1 });

  // insert second message as answer to first message
  const { data: [{ id: messageId2 }] } = await deep.insert({
    type_id: Message,
    string: { data: { value: 'second message' } },
    out: {
      data: {
        type_id: Reply,
        to_id: messageId1,
      }
    },
  });

  console.log({ messageId2 });

  const User = await deep.id('@deep-foundation/core', 'User'); // User type link id
  const ml1 = minilinks((await deep.select({
    type_id: { _in: [User, Reply, Message] },
  })).data);


  const f = link => (link.inByType[Reply] || []).map(r => [r.from.value.value, ...f(r.from)]);
  f(ml1.byId[deep.linkId]);


  const Tree = await deep.id('@deep-foundation/core', 'Tree');
  const TreeIncludeNode = await deep.id('@deep-foundation/core', 'TreeIncludeNode');
  const TreeIncludeUp = await deep.id('@deep-foundation/core', 'TreeIncludeUp');
  const TreeIncludeDown = await deep.id('@deep-foundation/core', 'TreeIncludeDown');
  const TreeIncludeCurrentFrom = await deep.id('@deep-foundation/core', 'TreeIncludeCurrentFrom');

  console.log({ Tree, TreeIncludeNode, TreeIncludeUp, TreeIncludeDown, TreeIncludeCurrentFrom });

  const { data: [{ id: messagingTree }] } = await deep.insert({
    type_id: Tree,
    in: {
      data: {
        // just for permission control
        type_id: Contain,
        from_id: deep.linkId,
        string: { data: { value: 'messagingTree' } },
      }
    },
    out: {
      data: [
        {
          type_id: TreeIncludeNode,
          to_id: User,
        },
        {
          type_id: TreeIncludeNode,
          to_id: Message,
        },
        {
          type_id: TreeIncludeUp,
          to_id: Reply,
        },
        {
          type_id: TreeIncludeUp,
          to_id: Author,
        },
        {
          type_id: TreeIncludeCurrentFrom,
          to_id: Join,
        },
        {
          type_id: TreeIncludeUp,
          to_id: await deep.id('@deep-foundation/core', 'Join'),
        },
      ]
    },
  });

  console.log({ messagingTree });

  const userRepliesTree = await deep.select({ _by_item: { group_id: { _eq: messagingTree }, path_item_id: { _eq: deep.linkId } } });
  userRepliesTree.data.length;


  const upperFromM2 = await deep.select({ type_id: { _neq: Reply }, _by_path_item: { group_id: { _eq: messagingTree }, item_id: { _eq: messageId2 } } });
  upperFromM2.data.length;
  const Rule = await deep.id('@deep-foundation/core', 'Rule');
  const RuleSubject = await deep.id('@deep-foundation/core', 'RuleSubject');
  const RuleObject = await deep.id('@deep-foundation/core', 'RuleObject');
  const RuleAction = await deep.id('@deep-foundation/core', 'RuleAction');
  const Selector = await deep.id('@deep-foundation/core', 'Selector');
  const SelectorInclude = await deep.id('@deep-foundation/core', 'SelectorInclude');
  const SelectorExclude = await deep.id('@deep-foundation/core', 'SelectorExclude');
  const SelectorTree = await deep.id('@deep-foundation/core', 'SelectorTree');
  const containTree = await deep.id('@deep-foundation/core', 'containTree');
  const AllowInsert = await deep.id('@deep-foundation/core', 'AllowInsert');
  const SelectorFilter = await deep.id('@deep-foundation/core', 'SelectorFilter');
  const BoolExp = await deep.id('@deep-foundation/core', 'BoolExp');
  const usersId = await deep.id('@deep-foundation/core', 'system', 'users');

  console.log({ Rule, RuleSubject, RuleObject, RuleAction, Selector, SelectorInclude, SelectorExclude, SelectorTree, containTree, AllowInsert, SelectorFilter, BoolExp, usersId });

  const replyInsertPermission = await deep.insert({
    type_id: Rule,
    out: {
      data: [
        {
          type_id: RuleSubject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: usersId,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: await deep.id('@deep-foundation/core', 'joinTree'),
                      },
                    },
                  },
                ]
              },
            },
          },
        },
        {
          type_id: RuleObject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: Reply,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                  {
                    type_id: SelectorFilter,
                    to: {
                      data: {
                        type_id: BoolExp,
                        object: {
                          data: {
                            value: {
                              // only if upper exists current user or join current user
                              to: {
                                _or: [
                                  {
                                    _by_item: { group_id: { _eq: messagingTree }, path_item_id: { _eq: 'X-Deep-User-Id' } },
                                  },
                                  {
                                    _by_item: { group_id: { _eq: messagingTree }, path_item: { out: { type_id: { _eq: Join }, to_id: { _eq: 'X-Deep-User-Id' } } } },
                                  },
                                ],
                              },
                              from: {
                                out: {
                                  type_id: { _eq: Author },
                                  to_id: { _eq: 'X-Deep-User-Id' },
                                },
                              },
                            },
                          },
                        },
                      },
                    },
                  },
                ],
              },
            },
          },
        },
        {
          type_id: RuleAction,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: AllowInsert,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                ],
              },
            },
          },
        },
      ],
    },
  });
  console.log({ replyInsertPermission: replyInsertPermission?.data?.[0] });

  // Ok then create message permission.
  const messagePermission = await deep.insert({
    type_id: Rule,
    out: {
      data: [
        {
          type_id: RuleSubject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: usersId,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: await deep.id('@deep-foundation/core', 'joinTree'),
                      },
                    },
                  },
                ]
              },
            },
          },
        },
        {
          type_id: RuleObject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: Message,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                ],
              },
            },
          },
        },
        {
          type_id: RuleAction,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: AllowInsert,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                ],
              },
            },
          },
        },
      ],
    },
  });

  console.log({ messagePermission: messagePermission?.data?.[0] });

  // Ok then create author permission.
  const authorPermission = await deep.insert({
    type_id: Rule,
    out: {
      data: [
        {
          type_id: RuleSubject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: usersId,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: await deep.id('@deep-foundation/core', 'joinTree'),
                      },
                    },
                  },
                ]
              },
            },
          },
        },
        {
          type_id: RuleObject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: Author,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                  {
                    type_id: SelectorFilter,
                    to: {
                      data: {
                        type_id: BoolExp,
                        object: {
                          data: {
                            value: {
                              to_id: { _eq: 'X-Deep-User-Id' },
                              from: {
                                _not: {
                                  out: {
                                    id: { _neq: 'X-Deep-Item-Id' },
                                    type_id: { _eq: Author },
                                  },
                                }
                              },
                            },
                          },
                        },
                      },
                    },
                  },
                ],
              },
            },
          },
        },
        {
          type_id: RuleAction,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: AllowInsert,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                ],
              },
            },
          },
        },
      ],
    },
  });

  console.log({ authorPermission: authorPermission?.data?.[0] });

  const joinPermission = await deep.insert({
    type_id: Rule,
    out: {
      data: [
        {
          type_id: RuleSubject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: usersId,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: await deep.id('@deep-foundation/core', 'joinTree'),
                      },
                    },
                  },
                ]
              },
            },
          },
        },
        {
          type_id: RuleObject,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: Join,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                  {
                    type_id: SelectorFilter,
                    to: {
                      data: {
                        type_id: BoolExp,
                        object: {
                          data: {
                            value: {
                              // only if owner of all reply tree can add join
                              from: {
                                _by_item: { group_id: { _eq: messagingTree }, path_item_id: { _eq: 'X-Deep-User-Id' } },
                              },
                            },
                          },
                        },
                      },
                    },
                  },
                ],
              },
            },
          },
        },
        {
          type_id: RuleAction,
          to: {
            data: {
              type_id: Selector,
              out: {
                data: [
                  {
                    type_id: SelectorInclude,
                    to_id: AllowInsert,
                    out: {
                      data: {
                        type_id: SelectorTree,
                        to_id: containTree,
                      },
                    },
                  },
                ],
              },
            },
          },
        },
      ],
    },
  });
  
  console.log({ joinPermission: joinPermission?.data?.[0] });

  const guest1 = await deep.guest();
  const deepGuest1 = new DeepClient({ deep, ...guest1 });

  const { data: [{ id: guest1Message1 }] } = await deepGuest1.insert({
    type_id: Message,
    string: { data: { value: 'guest1message1' } },
  });
  const { data: [{ id: guest1Message1Author }] } = await deepGuest1.insert({
    type_id: Author,
    from_id: guest1Message1,
    to_id: deepGuest1.linkId,
  });
  const { data: [{ id: guest1Message1Reply }] } = await deepGuest1.insert({
    type_id: Reply,
    from_id: guest1Message1,
    to_id: deepGuest1.linkId,
  });
  const { data: [{ id: guest1Message1Join }] } = await deepGuest1.insert({
    type_id: Join,
    from_id: guest1Message1Reply,
    to_id: deepGuest1.linkId,
  });

  console.log({ guest1Message1, guest1Message1Author, guest1Message1Reply, guest1Message1Join });

  await deepGuest1.delete(guest1Message1Reply);
  // you cant insert reply without author from message
  const { error: guest1Message2Error } = await deepGuest1.insert({
    type_id: Reply,
    from: {
      data: {
        type_id: Message,
        string: { data: { value: 'guest1message2' } },
      }
    },
    to_id: deepGuest1.linkId,
  }, { silent: true });
  const { data: [{ id: guest1Message2 }] } = await deepGuest1.insert({
    type_id: Message,
    string: { data: { value: 'guest1message2' } },
    out: {
      data: {
        type_id: Author,
        to_id: deepGuest1.linkId
      }
    },
  }, { silent: true });
  await deepGuest1.insert({
    type_id: Reply,
    to_id: deepGuest1.linkId,
    from_id: guest1Message2
  }, { silent: true });
  // Ok, next AllowSelect
  // For now, guest1 cant see guest1message2
  await deepGuest1.select(455);

  console.log('done');
};

f();